use clap::Parser;
use easymacros::x11_safe_wrapper::{Keysym, string_to_keysym, XDisplay};
use std::ffi::CString;
use std::process::{Command, exit};
use std::time::Duration;
use std::{fs, thread};
use std::io::stdin;
use x11::keysym::XK_Shift_L;
use easymacros::chartbl::CHARTBL;

/// Macro player module for easymacros. It's partially compatible with xmacro macros, with aim for full compatibility.
#[derive(Parser, Debug)]
#[clap(author, version, about, long_about = None)]
struct Args {
    /// The file that contains the macro to run.
    #[clap(value_parser, value_name = "input_file", value_hint = clap::ValueHint::FilePath)]
    input_file: Option<std::path::PathBuf>,
    /// Display to run the macro on. This uses the $DISPLAY environment variable by default.
    #[clap(short, long)]
    display: Option<String>,
}

fn main() {
    let args = Args::parse();

    let display = get_remote(args.display);

    if let Some(input_file_path) = args.input_file {
        let input_file_contents = fs::read_to_string(input_file_path)
            .expect("Couldn't read macro file");

        for instruction in input_file_contents.lines() {
            run_instruction(instruction, &display);
        }
    } else {
        let stdin = stdin();

        loop {
            let mut line = String::new();
            stdin.read_line(&mut line).expect("Couldn't read line from stdin");
            // Without this it crashes because apparently it doesn't properly read the next input line?
            println!();
            line = line.trim().to_string();
            run_instruction(&*line, &display);
        }
    }



    display.close();
}

fn get_remote(display_name: Option<String>) -> XDisplay {
    let display = XDisplay::open(display_name);

    if !display.has_xtest() {
        eprintln!("XTest not supported!");
        display.close();
        exit(1)
    }

    display.grab_control();
    display.sync();

    display
}

fn run_instruction(instruction: &str, dpy: &XDisplay) {
    let instruction_split: Vec<&str> = instruction.split(' ').collect();

    match instruction_split[0] {
        "Delay" => thread::sleep(Duration::from_millis(instruction_split[1].parse().unwrap())),
        "ButtonPress" => dpy.send_fake_buttonpress(instruction_split[1].parse().unwrap()),
        "ButtonRelease" => dpy.send_fake_buttonrelease(instruction_split[1].parse().unwrap()),
        "MotionNotify" => dpy
            .send_fake_motion_event(instruction_split[1].parse().unwrap(), instruction_split[2].parse().unwrap()),
        "KeyCodePress" => dpy.send_fake_keypress_from_code(instruction_split[1].parse().unwrap()),
        "KeyCodeRelease" => dpy.send_fake_keyrelease_from_code(instruction_split[1].parse().unwrap()),
        "KeySymPress" => dpy.send_fake_keypress_from_keysym(instruction_split[1].parse().unwrap()),
        "KeySymRelease" => {
            dpy.send_fake_keyrelease_from_keysym(instruction_split[1].parse().unwrap())
        }
        "KeySym" => {
            let key: Keysym = instruction_split[1].parse().unwrap();
            dpy.send_fake_keypress_from_keysym(key);
            dpy.send_fake_keyrelease_from_keysym(key);
        }
        "KeyStrPress" => {
            dpy.send_fake_keypress_from_string(CString::new(instruction_split[1]).unwrap().as_bytes())
        }
        "KeyStrRelease" => dpy
            .send_fake_keyrelease_from_string(CString::new(instruction_split[1]).unwrap().as_bytes()),
        "KeyStr" => {
            let keystring = CString::new(instruction_split[1]).unwrap();
            dpy.send_fake_keypress_from_string(keystring.as_bytes());
            dpy.send_fake_keyrelease_from_string(keystring.as_bytes());
        }
        "String" => {
            for c in instruction["String".len()+1..].chars() {
                send_char(dpy, c);
            }
        }
        "ExecBlock" | "ExecNoBlock" => {
            let mut command = Command::new(instruction_split[1]);
            for arg in &instruction_split[2..] {
                command.arg(arg);
            }
            if instruction_split[0] == "ExecBlock" {
                command.status();
            } else {
                command.spawn();
            }
        }
        c => {
            panic!("Unknown command {:?}", instruction_split)
        }
    }
}

fn send_char(dpy: &XDisplay, c: char) {
    let keysym = string_to_keysym(CHARTBL[c as usize].as_ref());
    let keycode = dpy.keysym_to_keycode(keysym);

    if keycode == 0 {
        eprintln!("No keycode found for character '{}'", c);
        return;
    }

    let map_ks = dpy.get_keyboard_mapping(keycode, 1);
    if unsafe { map_ks[0] } == 0 {
        eprintln!("XGetKeyboardMapping failed (keycode: {})", keycode);
        return;
    }

    let (ks_lower, ks_upper) = dpy.convert_case(keysym);

    let mut shift_needed = true;
    if keysym == map_ks[0] && (keysym == ks_lower && keysym == ks_upper) {
        shift_needed = false;
    }
    if keysym == ks_lower && keysym != ks_upper {
        shift_needed = false;
    }

    if shift_needed { dpy.send_fake_keypress_from_keysym(XK_Shift_L as Keysym); }
    dpy.send_fake_keypress_from_code(keycode);
    dpy.send_fake_keyrelease_from_code(keycode);
    if shift_needed { dpy.send_fake_keyrelease_from_keysym(XK_Shift_L as Keysym); }
}